package org.eclipse.jdt.internal.compiler.lookup;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.util.List;
import java.util.Set;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.CaptureBinding;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.InferenceContext;
import org.eclipse.jdt.internal.compiler.lookup.InferenceVariable;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.PackageBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.lookup.UnresolvedReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.WildcardBinding;

public final class ArrayBinding extends TypeBinding {
  public static final FieldBinding ArrayLength = new FieldBinding(TypeConstants.LENGTH, (TypeBinding)TypeBinding.INT, 17, null, Constant.NotAConstant);
  
  public TypeBinding leafComponentType;
  
  public int dimensions;
  
  LookupEnvironment environment;
  
  char[] constantPoolName;
  
  char[] genericTypeSignature;
  
  public long[] nullTagBitsPerDimension;
  
  public ArrayBinding(TypeBinding type, int dimensions, LookupEnvironment environment) {
    this.tagBits |= 0x1L;
    this.leafComponentType = type;
    this.dimensions = dimensions;
    this.environment = environment;
    if (type instanceof UnresolvedReferenceBinding) {
      ((UnresolvedReferenceBinding)type).addWrapper(this, environment);
    } else {
      this.tagBits |= type.tagBits & 0x2000000060000880L;
    } 
    long mask = type.tagBits & 0x180000000000000L;
    if (mask != 0L) {
      this.nullTagBitsPerDimension = new long[this.dimensions + 1];
      this.nullTagBitsPerDimension[this.dimensions] = mask;
      this.tagBits |= 0x100000L;
    } 
  }
  
  public TypeBinding closestMatch() {
    if (isValidBinding())
      return this; 
    TypeBinding leafClosestMatch = this.leafComponentType.closestMatch();
    if (leafClosestMatch == null)
      return null; 
    return this.environment.createArrayType(this.leafComponentType.closestMatch(), this.dimensions);
  }
  
  public List<TypeBinding> collectMissingTypes(List<TypeBinding> missingTypes) {
    if ((this.tagBits & 0x80L) != 0L)
      missingTypes = this.leafComponentType.collectMissingTypes(missingTypes); 
    return missingTypes;
  }
  
  public void collectSubstitutes(Scope scope, TypeBinding actualType, InferenceContext inferenceContext, int constraint) {
    int actualDim;
    if ((this.tagBits & 0x20000000L) == 0L)
      return; 
    if (actualType == TypeBinding.NULL || actualType.kind() == 65540)
      return; 
    switch (actualType.kind()) {
      case 68:
        actualDim = actualType.dimensions();
        if (actualDim == this.dimensions) {
          this.leafComponentType.collectSubstitutes(scope, actualType.leafComponentType(), inferenceContext, constraint);
          break;
        } 
        if (actualDim > this.dimensions) {
          org.eclipse.jdt.internal.compiler.lookup.ArrayBinding actualReducedType = this.environment.createArrayType(actualType.leafComponentType(), actualDim - this.dimensions);
          this.leafComponentType.collectSubstitutes(scope, actualReducedType, inferenceContext, constraint);
        } 
        break;
    } 
  }
  
  public boolean mentionsAny(TypeBinding[] parameters, int idx) {
    return this.leafComponentType.mentionsAny(parameters, idx);
  }
  
  void collectInferenceVariables(Set<InferenceVariable> variables) {
    this.leafComponentType.collectInferenceVariables(variables);
  }
  
  TypeBinding substituteInferenceVariable(InferenceVariable var, TypeBinding substituteType) {
    TypeBinding substitutedLeaf = this.leafComponentType.substituteInferenceVariable(var, substituteType);
    if (TypeBinding.notEquals(substitutedLeaf, this.leafComponentType))
      return this.environment.createArrayType(substitutedLeaf, this.dimensions, this.typeAnnotations); 
    return this;
  }
  
  public char[] computeUniqueKey(boolean isLeaf) {
    char[] brackets = new char[this.dimensions];
    for (int i = this.dimensions - 1; i >= 0; ) {
      brackets[i] = '[';
      i--;
    } 
    return CharOperation.concat(brackets, this.leafComponentType.computeUniqueKey(isLeaf));
  }
  
  public char[] constantPoolName() {
    if (this.constantPoolName != null)
      return this.constantPoolName; 
    char[] brackets = new char[this.dimensions];
    for (int i = this.dimensions - 1; i >= 0; ) {
      brackets[i] = '[';
      i--;
    } 
    return this.constantPoolName = CharOperation.concat(brackets, this.leafComponentType.signature());
  }
  
  public String debugName() {
    if (hasTypeAnnotations())
      return annotatedDebugName(); 
    StringBuffer brackets = new StringBuffer(this.dimensions * 2);
    for (int i = this.dimensions; --i >= 0;)
      brackets.append("[]"); 
    return String.valueOf(this.leafComponentType.debugName()) + brackets.toString();
  }
  
  public String annotatedDebugName() {
    StringBuffer brackets = new StringBuffer(this.dimensions * 2);
    brackets.append(this.leafComponentType.annotatedDebugName());
    brackets.append(' ');
    AnnotationBinding[] annotations = getTypeAnnotations();
    for (int i = 0, j = -1; i < this.dimensions; i++) {
      if (annotations != null) {
        if (i != 0)
          brackets.append(' '); 
        while (++j < annotations.length && annotations[j] != null) {
          brackets.append(annotations[j]);
          brackets.append(' ');
        } 
      } 
      brackets.append("[]");
    } 
    return brackets.toString();
  }
  
  public int dimensions() {
    return this.dimensions;
  }
  
  public TypeBinding elementsType() {
    if (this.dimensions == 1)
      return this.leafComponentType; 
    AnnotationBinding[] oldies = getTypeAnnotations();
    AnnotationBinding[] newbies = Binding.NO_ANNOTATIONS;
    for (int i = 0, length = (oldies == null) ? 0 : oldies.length; i < length; i++) {
      if (oldies[i] == null) {
        System.arraycopy(oldies, i + 1, newbies = new AnnotationBinding[length - i - 1], 0, length - i - 1);
        break;
      } 
    } 
    return this.environment.createArrayType(this.leafComponentType, this.dimensions - 1, newbies);
  }
  
  public TypeBinding erasure() {
    TypeBinding erasedType = this.leafComponentType.erasure();
    if (TypeBinding.notEquals(this.leafComponentType, erasedType))
      return this.environment.createArrayType(erasedType, this.dimensions); 
    return this;
  }
  
  public LookupEnvironment environment() {
    return this.environment;
  }
  
  public char[] genericTypeSignature() {
    if (this.genericTypeSignature == null) {
      char[] brackets = new char[this.dimensions];
      for (int i = this.dimensions - 1; i >= 0; ) {
        brackets[i] = '[';
        i--;
      } 
      this.genericTypeSignature = CharOperation.concat(brackets, this.leafComponentType.genericTypeSignature());
    } 
    return this.genericTypeSignature;
  }
  
  public PackageBinding getPackage() {
    return this.leafComponentType.getPackage();
  }
  
  public int hashCode() {
    return (this.leafComponentType == null) ? super.hashCode() : this.leafComponentType.hashCode();
  }
  
  public boolean isCompatibleWith(TypeBinding otherType, Scope captureScope) {
    org.eclipse.jdt.internal.compiler.lookup.ArrayBinding otherArray;
    CaptureBinding otherCapture;
    TypeBinding otherLowerBound;
    if (equalsEquals(this, otherType))
      return true; 
    switch (otherType.kind()) {
      case 68:
        otherArray = (org.eclipse.jdt.internal.compiler.lookup.ArrayBinding)otherType;
        if (otherArray.leafComponentType.isBaseType())
          return false; 
        if (this.dimensions == otherArray.dimensions)
          return this.leafComponentType.isCompatibleWith(otherArray.leafComponentType); 
        if (this.dimensions < otherArray.dimensions)
          return false; 
        break;
      case 132:
        return false;
      case 516:
      case 8196:
        return ((WildcardBinding)otherType).boundCheck(this);
      case 4100:
        otherCapture = (CaptureBinding)otherType;
        if (otherType.isCapture() && (otherLowerBound = otherCapture.lowerBound) != null) {
          if (!otherLowerBound.isArrayType())
            return false; 
          return isCompatibleWith(otherLowerBound, captureScope);
        } 
        return false;
    } 
    switch ((otherType.leafComponentType()).id) {
      case 1:
      case 36:
      case 37:
        return true;
    } 
    return false;
  }
  
  public boolean isSubtypeOf(TypeBinding otherType) {
    org.eclipse.jdt.internal.compiler.lookup.ArrayBinding otherArray;
    if (equalsEquals(this, otherType))
      return true; 
    switch (otherType.kind()) {
      case 68:
        otherArray = (org.eclipse.jdt.internal.compiler.lookup.ArrayBinding)otherType;
        if (otherArray.leafComponentType.isBaseType())
          return false; 
        if (this.dimensions == otherArray.dimensions)
          return this.leafComponentType.isSubtypeOf(otherArray.leafComponentType); 
        if (this.dimensions < otherArray.dimensions)
          return false; 
        break;
      case 132:
        return false;
    } 
    switch ((otherType.leafComponentType()).id) {
      case 1:
      case 36:
      case 37:
        return true;
    } 
    return false;
  }
  
  public boolean isProperType(boolean admitCapture18) {
    return this.leafComponentType.isProperType(admitCapture18);
  }
  
  public int kind() {
    return 68;
  }
  
  public TypeBinding leafComponentType() {
    return this.leafComponentType;
  }
  
  public char[] nullAnnotatedReadableName(CompilerOptions options, boolean shortNames) {
    if (this.nullTagBitsPerDimension == null)
      return shortNames ? shortReadableName() : readableName(); 
    char[][] brackets = new char[this.dimensions][];
    for (int i = 0; i < this.dimensions; i++) {
      if ((this.nullTagBitsPerDimension[i] & 0x180000000000000L) != 0L) {
        char[][] fqAnnotationName;
        if ((this.nullTagBitsPerDimension[i] & 0x100000000000000L) != 0L) {
          fqAnnotationName = options.nonNullAnnotationName;
        } else {
          fqAnnotationName = options.nullableAnnotationName;
        } 
        char[] annotationName = shortNames ? 
          fqAnnotationName[fqAnnotationName.length - 1] : 
          CharOperation.concatWith(fqAnnotationName, '.');
        brackets[i] = new char[annotationName.length + 3];
        brackets[i][0] = '@';
        System.arraycopy(annotationName, 0, brackets[i], 1, annotationName.length);
        brackets[i][annotationName.length + 1] = '[';
        brackets[i][annotationName.length + 2] = ']';
      } else {
        (new char[2])[0] = '[';
        (new char[2])[1] = ']';
        brackets[i] = new char[2];
      } 
    } 
    return CharOperation.concat(this.leafComponentType.nullAnnotatedReadableName(options, shortNames), 
        CharOperation.concatWith(brackets, ' '), 
        ' ');
  }
  
  public int problemId() {
    return this.leafComponentType.problemId();
  }
  
  public char[] qualifiedSourceName() {
    char[] brackets = new char[this.dimensions * 2];
    for (int i = this.dimensions * 2 - 1; i >= 0; i -= 2) {
      brackets[i] = ']';
      brackets[i - 1] = '[';
    } 
    return CharOperation.concat(this.leafComponentType.qualifiedSourceName(), brackets);
  }
  
  public char[] readableName() {
    char[] brackets = new char[this.dimensions * 2];
    for (int i = this.dimensions * 2 - 1; i >= 0; i -= 2) {
      brackets[i] = ']';
      brackets[i - 1] = '[';
    } 
    return CharOperation.concat(this.leafComponentType.readableName(), brackets);
  }
  
  public void setTypeAnnotations(AnnotationBinding[] annotations, boolean evalNullAnnotations) {
    this.tagBits |= 0x200000L;
    if (annotations == null || annotations.length == 0)
      return; 
    this.typeAnnotations = annotations;
    if (evalNullAnnotations) {
      long nullTagBits = 0L;
      if (this.nullTagBitsPerDimension == null)
        this.nullTagBitsPerDimension = new long[this.dimensions + 1]; 
      int dimension = 0;
      for (int i = 0, length = annotations.length; i < length; i++) {
        AnnotationBinding annotation = annotations[i];
        if (annotation != null) {
          if (annotation.type.hasNullBit(64)) {
            nullTagBits |= 0x80000000000000L;
            this.tagBits |= 0x100000L;
          } else if (annotation.type.hasNullBit(32)) {
            nullTagBits |= 0x100000000000000L;
            this.tagBits |= 0x100000L;
          } 
        } else {
          if (nullTagBits != 0L) {
            this.nullTagBitsPerDimension[dimension] = nullTagBits;
            nullTagBits = 0L;
          } 
          dimension++;
        } 
      } 
      this.tagBits |= this.nullTagBitsPerDimension[0];
    } 
  }
  
  public char[] shortReadableName() {
    char[] brackets = new char[this.dimensions * 2];
    for (int i = this.dimensions * 2 - 1; i >= 0; i -= 2) {
      brackets[i] = ']';
      brackets[i - 1] = '[';
    } 
    return CharOperation.concat(this.leafComponentType.shortReadableName(), brackets);
  }
  
  public char[] sourceName() {
    char[] brackets = new char[this.dimensions * 2];
    for (int i = this.dimensions * 2 - 1; i >= 0; i -= 2) {
      brackets[i] = ']';
      brackets[i - 1] = '[';
    } 
    return CharOperation.concat(this.leafComponentType.sourceName(), brackets);
  }
  
  public void swapUnresolved(UnresolvedReferenceBinding unresolvedType, ReferenceBinding resolvedType, LookupEnvironment env) {
    if (this.leafComponentType == unresolvedType) {
      this.leafComponentType = env.convertUnresolvedBinaryToRawType((TypeBinding)resolvedType);
      if (this.leafComponentType != resolvedType)
        this.id = (env.createArrayType(this.leafComponentType, this.dimensions, this.typeAnnotations)).id; 
      this.tagBits |= this.leafComponentType.tagBits & 0x2000000060000080L;
    } 
  }
  
  public String toString() {
    return (this.leafComponentType != null) ? debugName() : "NULL TYPE ARRAY";
  }
  
  public TypeBinding unannotated() {
    return hasTypeAnnotations() ? this.environment.getUnannotatedType(this) : this;
  }
  
  public TypeBinding withoutToplevelNullAnnotation() {
    if (!hasNullTypeAnnotations())
      return this; 
    AnnotationBinding[] newAnnotations = this.environment.filterNullTypeAnnotations(this.typeAnnotations);
    return this.environment.createArrayType(this.leafComponentType, this.dimensions, newAnnotations);
  }
  
  public TypeBinding uncapture(Scope scope) {
    if ((this.tagBits & 0x2000000000000000L) == 0L)
      return this; 
    TypeBinding leafType = this.leafComponentType.uncapture(scope);
    return scope.environment().createArrayType(leafType, this.dimensions, this.typeAnnotations);
  }
  
  public boolean acceptsNonNullDefault() {
    return true;
  }
  
  public long updateTagBits() {
    if (this.leafComponentType != null)
      this.tagBits |= this.leafComponentType.updateTagBits(); 
    return super.updateTagBits();
  }
}
